//Distributed under Kellen's sort-of-free license.
//http://people.colgate.edu/kmyers/license.html
//(c) Kellen Myers and Joseph Parrish 2007

// vim:set ts=4:

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "proc.h"
#include "gen.h"

int (*fn)(int *) = NULL;
static int fn_code_len = 0;

static char *eat_integer(char *start, char end, int *n)
{
	if (n) *n = atoi(start);
	start = strchr(start, end);
	if (start) return start + 1;
	return strchr(start, '\0');
}

// FIXME: handle overflow?  64-bit math?
void create_function(char *string, int *num_vars)
{
	char *cur, *eq_start;
	char *code, *cur_code;
	int num_terms, num_factors;
	int v, e, t, f;

	if (fn) free((void *)fn);

	cur = string;
	cur = eat_integer(cur, ']', num_vars);
	cur = eat_integer(cur, ';', &num_terms);
	eq_start = cur;

	fn_code_len = 0;
#define COUNT_CODE(x) { fn_code_len += x##_len; }
	COUNT_CODE(f_prelude);
	for (t = 0; t < num_terms; t++) {
		COUNT_CODE(f_term_init);
		cur = eat_integer(cur, ':', &num_factors);
		for (f = 0; f < num_factors; f++) {
			cur = eat_integer(cur, '.', &v);
			cur = eat_integer(cur, '*', &e);
			if (v) {
				COUNT_CODE(f_term_get_var);
				while (e) {
					COUNT_CODE(f_mul);
					e--;
				}
			} else {
				COUNT_CODE(f_term_set_constant);
				COUNT_CODE(f_mul);
			}
		}
		COUNT_CODE(f_add);
		cur = eat_integer(cur, ',', NULL);
	}
	COUNT_CODE(f_return);

	code = (char *)malloc(fn_code_len);
	cur_code = code;
	cur = eq_start;

#define ADD_TO_CODE(x) { memcpy(cur_code, x, x##_len); cur_code += x##_len; }
	ADD_TO_CODE(f_prelude);
	for (t = 0; t < num_terms; t++) {
		ADD_TO_CODE(f_term_init);
		cur = eat_integer(cur, ':', &num_factors);
		for (f = 0; f < num_factors; f++) {
			cur = eat_integer(cur, '.', &v);
			cur = eat_integer(cur, '*', &e);
			if (v) {
				f_fill_var_num(v);
				// FIXME: division unsupported.
				if (e < 0) abort();
				ADD_TO_CODE(f_term_get_var);
				while (e) {
					ADD_TO_CODE(f_mul);
					e--;
				}
			} else {
				f_fill_constant(e);
				ADD_TO_CODE(f_term_set_constant);
				ADD_TO_CODE(f_mul);
			}
		}
		ADD_TO_CODE(f_add);
		cur = eat_integer(cur, ',', NULL);
	}
	ADD_TO_CODE(f_return);

	fn = (int (*)(int *))code;
}

void dump_code()
{
	unsigned char *code = (unsigned char *)fn;
	int i;

	printf("Code dump: (%d bytes)\n", fn_code_len);
	printf("=======================\n");
	for (i = 0; i < fn_code_len; ) {
		printf("%02x ", code[i]);
		i++;
		if ((i % 8) == 0) printf("\n");
	}
	if ((i % 8) != 0) printf("\n");
	printf("=======================\n");
}

