//Distributed under Kellen's sort-of-free license.
//http://people.colgate.edu/kmyers/license.html
//(c) Kellen Myers and Joseph Parrish 2007

// vim:set ts=4:

#include <stdlib.h>
#include <stdio.h>

#include "gen.h"
#include "colorings.h"
#ifdef DEBUG_LOUD
# include "stats.h"
#endif
#include "rado.h"
#include "global.h"

// this allows us to profile the generated function if need be.
#ifdef DEBUG
int do_fn(int *var_array)
{
	return fn(var_array);
}
#else
# define do_fn(va) fn(va)
#endif

// returns true if the coloring is valid.
static int color_valid(coloring_t *c, iterator_t *iters, int *var_array, int num_vars, rado_info_t *ri)
{
	int held;
	ri->tries = 0;

	for (held = 0; held < num_vars; held++) {
		iter_reset(iters, num_vars, c);
		iter_hold(&iters[held]);
		do {
			ri->tries++;
			fill_vars(iters, var_array, num_vars);
			if (!do_fn(var_array)) {
				ri->outcome = 0;
				return 0; // not valid.
			}
		} while (next(c, iters, 0, num_vars));
	}

	ri->outcome = 1;
	return 1; // valid.
}

void rado(int fd, char *eq, coloring_t *c, int stop, rado_info_t *ri, int die_on_stop)
{
	iterator_t *iters;
	int num_vars, *var_array;
	int start, first_color;

	ri->max_valid_len = 0;
	ri->valid_at_max = 0;
	create_function(eq, &num_vars);

	var_array = (int *)malloc(sizeof(int) * num_vars);
	iters = (iterator_t *)malloc(sizeof(iterator_t) * num_vars);

	start = c->last_bit;
	ri->max_len = c->last_bit;
	first_color = color_read(c, start);
	ri->done = 0;

	while (color_read(c, start) == first_color) {
		color_valid(c, iters, var_array, num_vars, ri);
#ifdef DEBUG_LOUD
		color_dump(c);
		printf("Coloring %s, color len=%d, peak color len=%d.\n"
			"Elements in %s=%d, %d variable assignments, %.5f%% done.\n\n",
			ri->outcome ? "valid" : "invalid",
			c->last_bit, ri->max_len,
			color_read(c, c->last_bit) ? "blue" : "red",
			num_in_color(c), ri->tries,
			color_progress(c));
#ifdef WIN32
		fflush(stdout);
#endif
#endif
		// reply to all pending info requests.
		while (check_front(fd, c, ri)) {}
		if (status == 0) return;

		if (ri->outcome) {
			if (c->last_bit > ri->max_valid_len) {
				ri->valid_at_max = 0; // in effect, clear stored colorings.
				ri->max_valid_len = c->last_bit;
			}
			if (c->last_bit == ri->max_valid_len) {
				if (ri->valid_at_max == ri->colorings_allocd) {
					// need more space for colorings.
					int i;
					ri->colorings_allocd *= 2;
					ri->colors_at_max = (coloring_t *)realloc(ri->colors_at_max, sizeof(coloring_t) * ri->colorings_allocd);
					for (i = ri->colorings_allocd / 2; i < ri->colorings_allocd; i++) {
						ri->colors_at_max[i].bits = NULL;
					}
				}
				// store the coloring.
				copy_coloring(&ri->colors_at_max[ri->valid_at_max], c);
				ri->valid_at_max++;
			}
			if (c->last_bit == stop) {
				if (die_on_stop) return;
				color_backtrack(c);
			} else {
				color_grow(c);
				if (c->last_bit > ri->max_len) {
					ri->max_len = c->last_bit;
				}
			}
		} else {
			color_backtrack(c);
		}
		// prevent backtracking before start.
		if (c->last_bit < start) break;
	}

	free(var_array);
	free(iters);

	if (ri->max_valid_len != ri->max_len) ri->done = 1;
}

